package com.codenic.codenic_alarm

import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.app.Service
import android.content.Intent
import android.media.Ringtone
import android.media.RingtoneManager
import android.os.Binder
import android.os.IBinder
import android.os.Build
import android.util.Log
import androidx.core.app.NotificationCompat

/** A data class that represents an alarm. */
data class AlarmInfo(val id: Int, val title: String, val content: String) {
    /** Converts this alarm to a JSON object. */
    fun toJson(): Map<String, Any>  {
        return mapOf(
            "id" to id,
            "title" to title,
            "content" to content,
        )
    }
}

/** A listener that listens to events from the [AlarmService]. */
interface AlarmServiceListener {
    /**
     * Called when an alarm has been started.
     *
     * The [alarmInfo] parameter contains the details of the alarm that has been started.
     */
    fun onAlarmStarted(alarmInfo: AlarmInfo)

    /**
     * Called when an alarm has been dismissed.
     */
    fun onAlarmDismissed()
}

/**
 * A foreground service that plays the alarm ringtone and shows a notification to represent an
 * active alarm.
 *
 * The lifetime of the alarm is tied to the lifetime of this service. If this service is killed,
 * then the triggered alarm is also killed.
 */
class AlarmService: Service() {
    /** A [Binder] that is used to bind to this service. */
    inner class LocalBinder : Binder() {
        fun getService(): AlarmService = this@AlarmService
    }

    /** The action and arguments for triggering a start alarm intent of the service. */
    class StartAlarmIntent {
        companion object {
            const val ACTION = "action_start_alarm"

            const val EXTRA_ALARM_ID = "id"
            const val EXTRA_ALARM_TITLE = "title"
            const val EXTRA_ALARM_CONTENT = "content"
        }
    }

    /** The action and arguments for triggering a dismiss alarm intent of the service. */
    class DismissAlarmIntent {
        companion object {
            const val ACTION = "action_dismiss_alarm"
        }
    }

    companion object {
        /** The ID of the notification channel for the foreground service. */
        const val NOTIFICATION_CHANNEL_ID = "codenic_alarm"

        /** The ID of the foreground service. */
        const val FOREGROUND_SERVICE_ID = 42130820

        /** A reference to the running instance of this singleton service. */
        private var _instance: AlarmService? = null

        /** Returns the running instance of this singleton service. */
        fun getInstance(): AlarmService? = _instance
    }

    /** A [Binder] that is used to bind to this service. */
    private val binder = LocalBinder()

    /** The alarm that is currently being played. */
    var alarmInfo: AlarmInfo? = null

    /** The ringtone that is currently being played. */
    private var ringtone: Ringtone? = null

    /** A listener that listens to events from this service. */
    private var alarmServiceListener: AlarmServiceListener? = null

    /**
     * Sets the [alarmServiceListener] that listens to events from this service, such as when an
     * alarm has been started or dismissed.
     */
    fun setAlarmServiceListener(listener: AlarmServiceListener) {
        alarmServiceListener = listener
    }

    override fun onBind(intent: Intent?): IBinder = binder

    override fun onStartCommand(intent: Intent?, flags: Int, startId: Int): Int {
        when (intent?.action) {
            StartAlarmIntent.ACTION -> {
                val alarmId =
                    intent.getIntExtra(StartAlarmIntent.EXTRA_ALARM_ID, -1)
                val alarmTitle =
                    intent.getStringExtra(StartAlarmIntent.EXTRA_ALARM_TITLE)!!
                val alarmContent =
                    intent.getStringExtra(StartAlarmIntent.EXTRA_ALARM_CONTENT)!!

                alarmInfo = AlarmInfo(alarmId, alarmTitle, alarmContent)

                showAlarmNotification(alarmInfo!!)
                playRingtone()

                alarmServiceListener?.onAlarmStarted(alarmInfo!!)
            }
            DismissAlarmIntent.ACTION -> {
                alarmServiceListener?.onAlarmDismissed()

                // Terminate the service to stop the alarm
                stopSelf()
            }
        }

        return START_STICKY
    }

    override fun onCreate() {
        super.onCreate()

        _instance = this

        createNotificationChannel()
        setupRingtone()
    }

    override fun onDestroy() {
        stopRingtone()

        _instance = null

        super.onDestroy()
    }

    /** Configures the ring tone to be played by this service. */
    private fun setupRingtone() {
        val ringtoneUri =
            RingtoneManager.getActualDefaultRingtoneUri(this, RingtoneManager.TYPE_ALARM)

        if (ringtoneUri == null) {
            Log.w("AlarmService", "No ringtone found")
            return
        }

        ringtone = RingtoneManager.getRingtone(this, ringtoneUri)

        if (ringtone == null) {
            Log.w("AlarmService", "Cannot get ringtone")
            return
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.P) {
            ringtone?.volume = 1.0f
            ringtone?.isLooping = true
        }
    }

    /** Creates a notification channel for the foreground service. */
    private fun createNotificationChannel() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            val serviceChannel = NotificationChannel(
                NOTIFICATION_CHANNEL_ID,
                "Alarm",
                NotificationManager.IMPORTANCE_HIGH,
            )

            serviceChannel.setSound(null, null)

            val notificationManager = getSystemService(NotificationManager::class.java)
            notificationManager.createNotificationChannel(serviceChannel)
        }
    }

    /**
     * Shows a non-dismissible notification for the foreground service, while at the same time,
     * provides details and action buttons about the triggered alarm.
     */
    private fun showAlarmNotification(alarmInfo: AlarmInfo) {
        Log.d("AlarmService", "startForegroundService called")

        val launchAppIntent = this.packageManager.getLaunchIntentForPackage(this.packageName)
        val launchAppPendingIntent = PendingIntent.getActivity(
            this,
            0,
            launchAppIntent,
            PendingIntent.FLAG_IMMUTABLE,
        )

        val dismissAlarmIntent = Intent(this, AlarmService::class.java)
        dismissAlarmIntent.action = DismissAlarmIntent.ACTION

        val dismissAlarmPendingIntent = PendingIntent.getService(
            this,
            0,
            dismissAlarmIntent,
            PendingIntent.FLAG_IMMUTABLE,
        )

        val notificationIcon =
            resources.getIdentifier("ic_launcher", "drawable", this.packageName)

        val notification = NotificationCompat.Builder(this, NOTIFICATION_CHANNEL_ID)
            .setContentTitle(alarmInfo.title)
            .setContentText(alarmInfo.content)
            .setContentIntent(launchAppPendingIntent)
            .setSmallIcon(
                if (notificationIcon != 0) notificationIcon
                else R.drawable.notification_icon
            )
            .setPriority(NotificationCompat.PRIORITY_MAX)
            .setStyle(NotificationCompat.BigTextStyle())
            .setVisibility(NotificationCompat.VISIBILITY_PUBLIC)
            .setAutoCancel(true)
            .setOngoing(true)
            .addAction(R.drawable.action_open_icon, "Open", launchAppPendingIntent)
            .addAction(R.drawable.action_close_icon, "Dismiss", dismissAlarmPendingIntent)
            .build()

        startForeground(FOREGROUND_SERVICE_ID, notification)
    }

    /**
     * Plays the alarm [ringtone].
     *
     * If the ]ringtone is already playing, then this does nothing.
     */
    private fun playRingtone() {
        if (ringtone == null) {
            Log.w("AlarmService", "Cannot play ringtone because it is null")
            return
        }

        if (!ringtone!!.isPlaying()) {
            ringtone!!.play()
        }
    }

    /**
     * Stops the alarm [ringtone].
     *
     * If the ringtone is not playing, then this does nothing.
     */
    private fun stopRingtone() {
        if (ringtone == null) {
            return
        }

        ringtone!!.stop()
    }

}