// ignore_for_file: public_member_api_docs, sort_constructors_first
import 'dart:async';

/// An interface for scheduling alarms on Android and iOS.
abstract interface class CodenicAlarmInterface {
  /// Initializes the plugin.
  ///
  /// This must be called at the start of the app.
  ///
  /// Returns `true` if the plugin was successfully initialized or has already
  /// been initialized. Otherwise, `false` is returned..
  FutureOr<bool> initialize();

  /// Schedules an alarm to be executed once in the future after the specified
  /// [delay].
  FutureOr<bool> schedule({
    required Duration delay,
    required AlarmInfo alarmInfo,
  });

  /// Schedules an alarm to be executed once in the future at the specified
  /// [schedule].
  FutureOr<bool> scheduleAt({
    required DateTime schedule,
    required AlarmInfo alarmInfo,
  });

  /// Cancels the alarm with the specified [id].
  ///
  /// Returns `true` if the alarm was successfully canceled or the [id] does
  /// not reference an alarm. Otherwise, `false` is returned on failure.
  FutureOr<bool> cancel({required int id});

  /// Dismisses the currently triggered or active alarm.
  ///
  /// If there are no active alarms, this method will do nothing.
  FutureOr<void> dismissAlarm();

  /// Returns the currently active alarm, if any.
  ///
  /// This can be used to display the active alarm in the UI or check if an
  /// alarm is currently active.
  FutureOr<AlarmInfo?> activeAlarm();
}

/// {@template AlarmInfo}
/// A model containing information about an alarm.
/// {@endtemplate}
class AlarmInfo {
  /// {@macro AlarmInfo}
  const AlarmInfo({
    required this.id,
    required this.title,
    required this.content,
  });

  /// The unique identifier of the alarm.
  final int id;

  /// The title of the alarm.
  final String title;

  /// The content of the alarm.
  final String content;

  /// Converts the [AlarmInfo] to a JSON object.
  Map<String, dynamic> toJson() {
    return <String, dynamic>{
      'id': id,
      'title': title,
      'content': content,
    };
  }

  /// Creates an [AlarmInfo] from a JSON object.
  factory AlarmInfo.fromJson(Map<String, dynamic> json) {
    return AlarmInfo(
      id: json['id'] as int,
      title: json['title'] as String,
      content: json['content'] as String,
    );
  }
}
